from __future__ import annotations

import asyncio
import gzip
import logging
from hashlib import md5
from typing import Any

from aioredis import from_url
from aioredis.exceptions import ConnectionError

log = logging.getLogger(__name__)


def _md5_str(_str: str) -> str:
    return md5(_str.encode("utf-8"), usedforsecurity=False).hexdigest()


async def gen_key(val: str, *args: str) -> str:
    """ Generate a unique key from the given arguments. """
    loop = asyncio.get_running_loop()
    unique_key = await loop.run_in_executor(None, _md5_str, val)
    for arg in args:
        unique_key = unique_key + await loop.run_in_executor(None,
                                                             _md5_str,
                                                             arg)
    unique_key = await loop.run_in_executor(None, _md5_str,
                                            unique_key)
    return unique_key


class Cache(object):
    """ Wrapper around aioredis to manage cache. """

    def __init__(self, uri: str, ttl: int):
        log.info("Initializing!")
        self.uri = uri
        self.ttl = ttl
        self.redis = from_url(
            self.uri,
            decode_responses=False
        )

    async def _is_online(self) -> bool:
        """ Check if the redis server is online. """
        try:
            await self.redis.ping()
        except ConnectionError:
            log.error('cannot connect to redis: %s', self.uri)
            return False
        except Exception as ex:  # pylint: disable=broad-except
            log.exception(ex)
            return False
        return True

    async def get(self, key: str) -> Any | None:
        """ Get the content of the given key. """
        if not await self._is_online():
            return None

        async with self.redis as redis_client:
            cached_ok = await redis_client.get(key)
            if cached_ok:
                return gzip.decompress(cached_ok).decode("utf-8")
        return None

    async def set(self, key: str, content: str) -> None:
        """ Set the content of the given key. """
        if not await self._is_online():
            return None

        async with self.redis as redis_client:
            await redis_client.set(
                key,
                gzip.compress(content.encode("utf-8")),
                ex=self.ttl
            )
        return None



#!/usr/bin/env python3
import asyncio
import logging
import signal

import grpc

from .protobuf.youtube_dl_tiny_grpc_pb2_grpc import \
    add_YoutubeDLServicer_to_server as AddYoutubeDLServer
from .util import ProcessPoolExecutor
from .cache import Cache
from .youtube_dl_service import YoutubeDLServer, \
    configure as configure_youtube_dl_server, \
    shutdown_pool as shutdown_youtube_dl_server

log = logging.getLogger(__name__)


class Server:
    """ YoutubeDL's gRPC Server """

    # Holds the shutdown method for the caller to call and clean up
    # eg: loop.run_until_complete(*server.cleanup)
    cleanup = []
    not_in_shutdown = True  # This triggers an ordered shutdown of the server
    compression_algorithms = {
        "none":    grpc.Compression.NoCompression,
        "deflate": grpc.Compression.Deflate,
        "gzip":    grpc.Compression.Gzip,
    }

    def __init__(self, args: dict):
        log.info("Initializing!")

        base_youtube_dl_args = {
            'verbose':   args['youtube_dl_verbose'] if "youtube_dl_verbose" in args else False,
            'quiet':     args['youtube_dl_no_quiet'] if "youtube_dl_no_quiet" in args else False,
            'simulate':  True,
            'call_home': False
        }

        if "youtube_dl_cookies_file" in args and \
                args['youtube_dl_cookies_file'] is not None and \
                args['youtube_dl_cookies_file'] != "":
            base_youtube_dl_args['cookiefile'] = \
                args['youtube_dl_cookies_file']

        redis = None
        if "redis_enable" in args and \
                args['redis_enable']:
            redis = Cache(args['redis_uri'], args['redis_ttl'])

        proxy_list = None
        if "youtube_dl_proxy_list" in args and args['youtube_dl_proxy_list'] is not None and \
                args['youtube_dl_proxy_list'] != "":
            proxy_list = args['youtube_dl_proxy_list'].split(',')

        configure_youtube_dl_server(
            base_youtube_dl_args,
            ProcessPoolExecutor(
                max_workers=args['youtube_dl_max_workers'] if "youtube_dl_max_workers" in args else 1
            ),
            redis,
            proxy_list
        )

        self.grpc_graceful_shutdown_timeout = \
            args['grpc_graceful_shutdown_timeout'] if "grpc_graceful_shutdown_timeout" in args else 0

        self.server = grpc.aio.server(
            compression=self.compression_algorithms[
                args['grpc_compression_algorithm'] if "grpc_compression_algorithm" in args else "none"
            ]
        )

        AddYoutubeDLServer(YoutubeDLServer(), self.server)

        if "grpc_no_reflection" in args and not args['grpc_no_reflection']:
            # pylint: disable=import-outside-toplevel
            from grpc_reflection.v1alpha import reflection

            from .protobuf.youtube_dl_tiny_grpc_pb2 import \
                DESCRIPTOR
            reflection.enable_server_reflection((
                DESCRIPTOR.services_by_name['YoutubeDL'].full_name,
                reflection.SERVICE_NAME,
            ), self.server)

        self.server.add_insecure_port(f"[::]:{args['grpc_port']}")

        # Add our method for the caller to clean up
        self.cleanup.append(self.shutdown())

        # Add signal handler for graceful shutdown
        signal.signal(signal.SIGINT, self._exit_gracefully)
        signal.signal(signal.SIGTERM, self._exit_gracefully)

    def _exit_gracefully(self, signum, frame):
        """ Internal method to exit gracefully on SIGINT or SIGTERM """
        if self.not_in_shutdown:
            log.info("starting graceful shutdown")
            self.not_in_shutdown = False

    async def shutdown(self) -> None:
        """ Shutdown the server """
        self._exit_gracefully(None, None)
        log.info("stopping server")
        await self.server.stop(self.grpc_graceful_shutdown_timeout)
        log.info("stopping youtube_dl process pool")
        shutdown_youtube_dl_server()
        log.info("waiting for server to stop")
        await self.server.wait_for_termination(self.grpc_graceful_shutdown_timeout)

    async def run(self) -> None:
        """ Run the server """
        await self.server.start()

        log.info("server started")
        # Sleep until we get SIGINT or SIGTERM...etc
        while self.not_in_shutdown:
            await asyncio.sleep(5)



#!/usr/bin/env python3
from google.protobuf.json_format import ParseDict, MessageToDict
import grpc
import os
import asyncio
import sys
import logging
import socket
from contextlib import closing

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from youtube_dl_tiny_grpc import (Server)
from youtube_dl_tiny_grpc.protobuf import youtube_dl_tiny_grpc_pb2_grpc
from youtube_dl_tiny_grpc.protobuf import youtube_dl_tiny_grpc_pb2

TEST_MATRIX = [
    {
        "meta": "Youtube Channel ID",
        "message": ParseDict(
            message=youtube_dl_tiny_grpc_pb2.ExtractInfoRequest(),
            js_dict={
                "url":     "https://www.youtube.com/channel/UCZJbZ-0Dxdj5ETtJqcJQO2g",
                "options": {"playlistend": 2}
            }),
        "expected_responses": 2
    },
    {
        "meta": "Youtube Playlist ID",
        "message": ParseDict(
            message=youtube_dl_tiny_grpc_pb2.ExtractInfoRequest(),
            js_dict={
                "url":     "https://www.youtube.com/playlist?list=PLbpi6ZahtOH7WFEYxB4kUB2QwsshALAwy",
                "options": {"playlistend": 2}
            }),
        "expected_responses": 2
    },
    {
        "meta": "Youtube Channel Short",
        "message": ParseDict(
            message=youtube_dl_tiny_grpc_pb2.ExtractInfoRequest(),
            js_dict={
                "url":     "https://www.youtube.com/c/unboxtherapy",
                "options": {"playlistend": 2}
            }),
        "expected_responses": 2
    },
]


async def run_tests(port: int) -> None:
    async with grpc.aio.insecure_channel(f"localhost:{port}") as channel:
        stub = youtube_dl_tiny_grpc_pb2_grpc.YoutubeDLStub(channel)
        for test_case in TEST_MATRIX:
            response_count = 0

            async for response in stub.ExtractInfo(test_case['message']):
                response_count = response_count + 1
                response_dict = MessageToDict(response)
                assert len(response_dict['requestedFormats']) > 0
                for format in response_dict['requestedFormats']:
                    assert len(format['url']) > 0

            try:
                assert response_count == test_case["expected_responses"]
            except AssertionError:
                print(
                    f"{test_case['meta']} failed: {response_count} != {test_case['expected_responses']}")
                raise


def run():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        loop = asyncio.get_event_loop()
        server = Server({"grpc_port": s.getsockname()[1]})
        server_task = loop.create_task(server.run())
        loop.run_until_complete(run_tests(s.getsockname()[1]))
        server_task.cancel()
        loop.stop()


if __name__ == '__main__':
    logging.basicConfig(level='DEBUG')
    run()



import signal

from concurrent.futures import ProcessPoolExecutor as PoolExecutor


def port_is_in_use(port: int) -> bool:
    import socket
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


class ProcessPoolExecutor(PoolExecutor):
    """ A ProcessPoolExecutor that ignores SIGINT signals. """

    def __init__(self, *args, **kwargs):
        kwargs['initializer'] = ProcessPoolExecutor.initializer
        super().__init__(*args, **kwargs)

    @staticmethod
    def initializer():
        signal.signal(signal.SIGINT, signal.SIG_IGN)



#!/usr/bin/env python3
# $ python -m youtube_dl_tiny_grpc

import sys

if __package__ is None and not hasattr(sys, 'frozen'):
    # direct call of __main__.py
    import os.path
    path = os.path.realpath(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(path)))

# pylint: disable=wrong-import-position
import youtube_dl_tiny_grpc

if __name__ == '__main__':
    youtube_dl_tiny_grpc.main()



from prometheus_client import CollectorRegistry
import glob
import os


_REGISTRY = CollectorRegistry()


def prepare_metrics_directory(d: str) -> None:
    os.environ["PROMETHEUS_MULTIPROC_DIR"] = d
    if not os.path.exists(d):
        os.mkdir(d)
    files = glob.glob(d + "/*")
    for f in files:
        os.remove(f)



#!/usr/bin/env python3
# coding: utf-8

from setproctitle import setproctitle
import logging
import asyncio
import errno
import os

from prometheus_client import multiprocess, start_http_server

from .metrics import _REGISTRY, prepare_metrics_directory
from .options import parse_opts
from .util import port_is_in_use
from .youtube_dl_tiny_grpc import (Server)

_PROGNAME = 'youtube-dl-tiny-grpc'
_LOG_FORMAT = '%(levelname)s - %(asctime)s - %(name)s - %(process)d - %(' \
              'message)s'


def main(argv=None):
    setproctitle(_PROGNAME)
    args = parse_opts(argv)

    if args.__dict__['debug']:
        logging.basicConfig(level='DEBUG', format=_LOG_FORMAT)
    elif args.__dict__['verbose']:
        logging.basicConfig(level='INFO', format=_LOG_FORMAT)
    else:
        logging.basicConfig()

    loop = asyncio.get_event_loop()

    if args.prometheus_enable:
        if port_is_in_use(args.prometheus_port):
            raise Exception(os.strerror(errno.EADDRINUSE))

        prepare_metrics_directory(args.prometheus_directory)

        multiprocess.MultiProcessCollector(_REGISTRY)
        start_http_server(args.prometheus_port, registry=_REGISTRY)

    if port_is_in_use(args.grpc_port):
        raise Exception(os.strerror(errno.EADDRINUSE))

    server = Server(args.__dict__)
    try:
        loop.run_until_complete(server.run())
    except:  # noqa: E722
        pass
    finally:
        loop.run_until_complete(*server.cleanup)
        loop.stop()


__all__ = ['main']



import argparse
from multiprocessing import cpu_count
import os

from setproctitle import getproctitle

from .version import __version__

_COMPRESSION_ALGORITHMS = ["none", "deflate", "gzip"]
_DESCRIPTION = 'A tiny slice of youtube-dl exposed over gRPC'


def parse_opts(argv: list) -> argparse.Namespace:
    """ Parse command line options. """

    parser = argparse.ArgumentParser(
        description=_DESCRIPTION,
        prog=getproctitle(),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    prometheus = parser.add_argument_group("prometheus")

    prometheus.add_argument(
        '--prometheus-port',
        default=os.getenv("PROMETHEUS_PORT", 8080),
        type=int,
        help='Port to listen on',
    )

    prometheus.add_argument(
        '--prometheus-enable',
        default=os.getenv("PROMETHEUS_ENABLE", False),
        action='store_true',
        help='Enable Prometheus metrics',
    )

    prometheus.add_argument(
        '--prometheus-directory',
        default=os.getenv("PROMETHEUS_DIRECTORY", "/tmp/youtube_dl_tiny_grpc_prometheus"),
        type=str,
        help='Prometheus directory to share between processes',
    )

    grpc = parser.add_argument_group("grpc")

    grpc.add_argument(
        '--grpc-port',
        default=os.getenv("GRPC_PORT", 50051),
        type=int,
        help='Port to listen on',
    )

    grpc.add_argument(
        '--grpc-graceful-shutdown-timeout',
        default=os.getenv("GRPC_GRACEFUL_SHUTDOWN_TIMEOUT", 60),
        type=int,
        help='Graceful shutdown timeout',
    )

    grpc.add_argument(
        '--grpc-no-reflection',
        default=os.getenv("GRPC_NO_REFLECTION", False),
        help='Disable gRPC server reflection',
    )

    grpc.add_argument(
        '--grpc-compression-algorithm',
        default=os.getenv("GRPC_COMPRESSION_ALGORITHM", "gzip"),
        choices=_COMPRESSION_ALGORITHMS,
        help='Compression algorithm for the server',
    )

    youtube_dl = parser.add_argument_group("youtube-dl")

    youtube_dl.add_argument(
        '--youtube-dl-max-workers',
        default=os.getenv("YOUTUBE_DL_MAX_WORKERS", cpu_count()),
        type=int,
        help='Max number of workers to use for youtube-dl',
    )

    youtube_dl.add_argument(
        '--youtube-dl-verbose',
        default=os.getenv("YOUTUBE_DL_VERBOSE", False),
        action='store_true',
        help='Verbose output for youtube-dl',
    )

    youtube_dl.add_argument(
        '--youtube-dl-no-quiet',
        default=os.getenv("YOUTUBE_DL_NO_QUIET", True),
        action='store_false',
        help='Disable quiet output for youtube-dl',
    )

    youtube_dl.add_argument(
        '--youtube-dl-proxy-list',
        default=os.getenv("YOUTUBE_DL_PROXY_LIST", ""),
        type=str,
        help="Comma separated list of proxies to use. \
        For example 'socks5://127.0.0.1:1080,http://127.0.0.1:8080'"
    )

    youtube_dl.add_argument(
        '--youtube-dl-cookies-file',
        default=os.getenv("YOUTUBE_DL_COOKIES_FILE", ""),
        type=str,
        help="File to read cookies from and dump cookie jar in"
    )

    redis = parser.add_argument_group("redis")

    redis.add_argument(
        '--redis-enable',
        action='store_true',
        default=os.getenv("REDIS_ENABLE", False),
        help='Enable Redis cache support',
    )

    redis.add_argument(
        '--redis-uri',
        type=str,
        default=os.getenv("REDIS_URI", 'redis://localhost:6379'),
        help='Redis URI to connect to',
    )

    redis.add_argument(
        '--redis-ttl',
        type=int,
        default=os.getenv("REDIS_TTL", 3600),
        help='TTL for cached results',
    )

    general = parser.add_argument_group("general")

    general.add_argument(
        '--version',
        action='version',
        version=f'%(prog)s {__version__}'
    )

    general.add_argument(
        '--debug',
        action='store_true',
        default=os.getenv("DEBUG", False),
        help='Enable debug logging',
    )

    general.add_argument(
        '--verbose',
        action='store_true',
        default=os.getenv("VERBOSE", False),
        help='Enable verbose logging',
    )

    args = parser.parse_args(argv)

    if args.verbose or args.debug is not None:
        print("configuration:")
        for arg in vars(args):
            print(f"  {arg}: {getattr(args, arg)}")

    return args



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: youtube_dl_tiny_grpc.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1ayoutube_dl_tiny_grpc.proto\"\xb7\t\n\x13\x45xtractInfoResponse\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05title\x18\x02 \x01(\t\x12-\n\x07\x66ormats\x18\x03 \x03(\x0b\x32\x1c.ExtractInfoResponse.Formats\x12\x33\n\nthumbnails\x18\x04 \x03(\x0b\x32\x1f.ExtractInfoResponse.Thumbnails\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x13\n\x0bupload_date\x18\x06 \x01(\t\x12\x10\n\x08uploader\x18\x07 \x01(\t\x12\x13\n\x0buploader_id\x18\x08 \x01(\t\x12\x14\n\x0cuploader_url\x18\t \x01(\t\x12\x12\n\nchannel_id\x18\n \x01(\t\x12\x13\n\x0b\x63hannel_url\x18\x0b \x01(\t\x12\x10\n\x08\x64uration\x18\x0c \x01(\r\x12\x12\n\nview_count\x18\r \x01(\r\x12\x12\n\ncategories\x18\x11 \x03(\t\x12\x0c\n\x04tags\x18\x12 \x03(\t\x12\x0f\n\x07\x63hannel\x18\x16 \x01(\t\x12\x11\n\textractor\x18\x17 \x01(\t\x12\x11\n\tthumbnail\x18\x1c \x01(\t\x12\x12\n\ndisplay_id\x18\x1d \x01(\t\x12\x0e\n\x06\x66ormat\x18  \x01(\t\x12\x11\n\tformat_id\x18! \x01(\t\x12\r\n\x05width\x18\" \x01(\r\x12\x0e\n\x06height\x18# \x01(\r\x12\x12\n\nresolution\x18$ \x01(\r\x12\x0b\n\x03\x66ps\x18% \x01(\r\x12\x0e\n\x06vcodec\x18& \x01(\t\x12\x0b\n\x03vbr\x18\' \x01(\x01\x12\x17\n\x0fstretched_ratio\x18( \x01(\r\x12\x0e\n\x06\x61\x63odec\x18) \x01(\t\x12\x0b\n\x03\x61\x62r\x18* \x01(\x01\x12\x0b\n\x03\x65xt\x18+ \x01(\t\x12\x37\n\x11requested_formats\x18, \x03(\x0b\x32\x1c.ExtractInfoResponse.Formats\x12\x0b\n\x03\x61sr\x18- \x01(\r\x12\x10\n\x08\x66ilesize\x18. \x01(\r\x12\x13\n\x0b\x66ormat_note\x18/ \x01(\t\x12\x0f\n\x07quality\x18\x30 \x01(\r\x12\x0b\n\x03tbr\x18\x31 \x01(\x01\x12\x0b\n\x03url\x18\x32 \x01(\t\x12\x11\n\tcontainer\x18\x33 \x01(\t\x12\x10\n\x08protocol\x18\x34 \x01(\t\x1a\xa3\x02\n\x07\x46ormats\x12\x0b\n\x03\x61sr\x18\x01 \x01(\r\x12\x10\n\x08\x66ilesize\x18\x02 \x01(\r\x12\x11\n\tformat_id\x18\x03 \x01(\t\x12\x13\n\x0b\x66ormat_note\x18\x04 \x01(\t\x12\x0b\n\x03\x66ps\x18\x05 \x01(\r\x12\x0e\n\x06height\x18\x06 \x01(\r\x12\x0f\n\x07quality\x18\x07 \x01(\r\x12\x0b\n\x03tbr\x18\x08 \x01(\x01\x12\x0b\n\x03url\x18\t \x01(\t\x12\r\n\x05width\x18\n \x01(\r\x12\x0b\n\x03\x65xt\x18\x0b \x01(\t\x12\x0e\n\x06vcodec\x18\x0c \x01(\t\x12\x0e\n\x06\x61\x63odec\x18\r \x01(\t\x12\x0b\n\x03\x61\x62r\x18\x0e \x01(\x01\x12\x11\n\tcontainer\x18\x10 \x01(\t\x12\x0e\n\x06\x66ormat\x18\x11 \x01(\t\x12\x10\n\x08protocol\x18\x12 \x01(\t\x12\x0b\n\x03vbr\x18\x14 \x01(\x01\x1aX\n\nThumbnails\x12\x0e\n\x06height\x18\x01 \x01(\r\x12\x0b\n\x03url\x18\x02 \x01(\t\x12\r\n\x05width\x18\x03 \x01(\r\x12\x12\n\nresolution\x18\x04 \x01(\t\x12\n\n\x02id\x18\x05 \x01(\t\"7\n\x10YoutubeDLOptions\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x13\n\x0bplaylistend\x18\x02 \x01(\r\"E\n\x12\x45xtractInfoRequest\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\"\n\x07options\x18\x02 \x01(\x0b\x32\x11.YoutubeDLOptions2I\n\tYoutubeDL\x12<\n\x0b\x45xtractInfo\x12\x13.ExtractInfoRequest\x1a\x14.ExtractInfoResponse\"\x00\x30\x01\x62\x06proto3')



_EXTRACTINFORESPONSE = DESCRIPTOR.message_types_by_name['ExtractInfoResponse']
_EXTRACTINFORESPONSE_FORMATS = _EXTRACTINFORESPONSE.nested_types_by_name['Formats']
_EXTRACTINFORESPONSE_THUMBNAILS = _EXTRACTINFORESPONSE.nested_types_by_name['Thumbnails']
_YOUTUBEDLOPTIONS = DESCRIPTOR.message_types_by_name['YoutubeDLOptions']
_EXTRACTINFOREQUEST = DESCRIPTOR.message_types_by_name['ExtractInfoRequest']
ExtractInfoResponse = _reflection.GeneratedProtocolMessageType('ExtractInfoResponse', (_message.Message,), {

  'Formats' : _reflection.GeneratedProtocolMessageType('Formats', (_message.Message,), {
    'DESCRIPTOR' : _EXTRACTINFORESPONSE_FORMATS,
    '__module__' : 'youtube_dl_tiny_grpc_pb2'
    # @@protoc_insertion_point(class_scope:ExtractInfoResponse.Formats)
    })
  ,

  'Thumbnails' : _reflection.GeneratedProtocolMessageType('Thumbnails', (_message.Message,), {
    'DESCRIPTOR' : _EXTRACTINFORESPONSE_THUMBNAILS,
    '__module__' : 'youtube_dl_tiny_grpc_pb2'
    # @@protoc_insertion_point(class_scope:ExtractInfoResponse.Thumbnails)
    })
  ,
  'DESCRIPTOR' : _EXTRACTINFORESPONSE,
  '__module__' : 'youtube_dl_tiny_grpc_pb2'
  # @@protoc_insertion_point(class_scope:ExtractInfoResponse)
  })
_sym_db.RegisterMessage(ExtractInfoResponse)
_sym_db.RegisterMessage(ExtractInfoResponse.Formats)
_sym_db.RegisterMessage(ExtractInfoResponse.Thumbnails)

YoutubeDLOptions = _reflection.GeneratedProtocolMessageType('YoutubeDLOptions', (_message.Message,), {
  'DESCRIPTOR' : _YOUTUBEDLOPTIONS,
  '__module__' : 'youtube_dl_tiny_grpc_pb2'
  # @@protoc_insertion_point(class_scope:YoutubeDLOptions)
  })
_sym_db.RegisterMessage(YoutubeDLOptions)

ExtractInfoRequest = _reflection.GeneratedProtocolMessageType('ExtractInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _EXTRACTINFOREQUEST,
  '__module__' : 'youtube_dl_tiny_grpc_pb2'
  # @@protoc_insertion_point(class_scope:ExtractInfoRequest)
  })
_sym_db.RegisterMessage(ExtractInfoRequest)

_YOUTUBEDL = DESCRIPTOR.services_by_name['YoutubeDL']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _EXTRACTINFORESPONSE._serialized_start=31
  _EXTRACTINFORESPONSE._serialized_end=1238
  _EXTRACTINFORESPONSE_FORMATS._serialized_start=857
  _EXTRACTINFORESPONSE_FORMATS._serialized_end=1148
  _EXTRACTINFORESPONSE_THUMBNAILS._serialized_start=1150
  _EXTRACTINFORESPONSE_THUMBNAILS._serialized_end=1238
  _YOUTUBEDLOPTIONS._serialized_start=1240
  _YOUTUBEDLOPTIONS._serialized_end=1295
  _EXTRACTINFOREQUEST._serialized_start=1297
  _EXTRACTINFOREQUEST._serialized_end=1366
  _YOUTUBEDL._serialized_start=1368
  _YOUTUBEDL._serialized_end=1441
# @@protoc_insertion_point(module_scope)



from __future__ import annotations
import asyncio
import json
import logging
import random

from google.protobuf.json_format import MessageToDict, ParseDict
from prometheus_client import Counter

import grpc
from youtube_dl import YoutubeDL

from .cache import Cache, gen_key
from .protobuf.youtube_dl_tiny_grpc_pb2 import ExtractInfoRequest, ExtractInfoResponse
from .protobuf.youtube_dl_tiny_grpc_pb2_grpc import (
    YoutubeDLServicer as YoutubeDLServerBase,
)
from .util import ProcessPoolExecutor

from .metrics import _REGISTRY

REQUEST_TOTAL = Counter('request_total', 'Total number of requests', registry=_REGISTRY)
CACHE_HIT_TOTAL = Counter('cache_hit_total', 'Total number of cache hits', registry=_REGISTRY)
CACHE_MISS_TOTAL = Counter('cache_miss_total', 'Total number of cache misses', registry=_REGISTRY)

# Offload all youtube_dl processing to a separate process in this pool
# Idea from https://github.com/grpc/grpc/issues/16001
_YOUTUBE_DL_PROCESS_POOL: ProcessPoolExecutor | None = None

# Default options for youtube-dl itself
_YOUTUBE_DL_DEFAULT_OPTS = {}

# The cache ssot
_YOUTUBE_DL_CACHE: Cache | None = None

# List of proxies to use in youtube-dl
_YOUTUBE_DL_PROXY_LIST: list | None = None

log = logging.getLogger(__name__)


def configure(
        default_opts: dict,
        process_pool: ProcessPoolExecutor,
        cache: Cache | None,
        proxy_list) -> None:
    global _YOUTUBE_DL_DEFAULT_OPTS
    global _YOUTUBE_DL_PROCESS_POOL
    global _YOUTUBE_DL_PROXY_LIST
    global _YOUTUBE_DL_CACHE
    _YOUTUBE_DL_DEFAULT_OPTS = default_opts
    _YOUTUBE_DL_PROCESS_POOL = process_pool
    _YOUTUBE_DL_PROXY_LIST = proxy_list
    _YOUTUBE_DL_CACHE = cache


def shutdown_pool() -> None:
    global _YOUTUBE_DL_PROCESS_POOL
    if isinstance(_YOUTUBE_DL_PROCESS_POOL, ProcessPoolExecutor):
        log.info("stopping process pool")
        _YOUTUBE_DL_PROCESS_POOL.shutdown(False)
    else:
        log.warning("cannot stop a non existent process pool")


class YoutubeDLServer(YoutubeDLServerBase):
    """Provides methods that implement functionality of YoutubeDL server."""

    def __init__(self):
        self.cache = _YOUTUBE_DL_CACHE

    @staticmethod
    def _extract_info(
            url: str,
            opts: dict = {},
            retries: int = 0,
            proxy: list = [],
            proxy_timeout: int = 30) -> dict:
        """Wrapper around youtube-dl's extract_info method
        to handle retries with proxy rotation."""
        count = retries
        real_ydl_opts = opts

        if proxy:
            r_proxy = random.choice(proxy)
            real_ydl_opts['proxy'] = r_proxy
            real_ydl_opts['socket_timeout'] = proxy_timeout

        try:
            ydl = YoutubeDL(real_ydl_opts)
            info = ydl.extract_info(url, False)
            return info
        except Exception as e:
            if count <= 0:
                raise e
            count = count - 1
            return YoutubeDLServer._extract_info(url, opts, count, proxy,
                                                 proxy_timeout)

    async def ExtractInfo(
            self,
            request: ExtractInfoRequest,
            context: grpc.aio.ServicerContext
    ) -> ExtractInfoResponse:
        if request.url == "" or not isinstance(request.url, str):
            await context.abort(grpc.StatusCode.INVALID_ARGUMENT,
                                "invalid URL")

        REQUEST_TOTAL.inc()

        ydl_custom_opts = MessageToDict(
            request.options,
            including_default_value_fields=False,
            preserving_proto_field_name=True
        )

        ydl_opts = {
            **_YOUTUBE_DL_DEFAULT_OPTS,
            **ydl_custom_opts,
        }

        info = None
        if self.cache is not None:
            cached_ok = await self.cache.get(
                await gen_key(request.url, json.dumps(ydl_opts))
            )
            if cached_ok:
                CACHE_HIT_TOTAL.inc()
                info = json.loads(cached_ok)

        if info is None:
            CACHE_MISS_TOTAL.inc()
            real_ydl_opts = ydl_opts

            loop = asyncio.get_event_loop()
            try:
                info = await loop.run_in_executor(
                    _YOUTUBE_DL_PROCESS_POOL,
                    self._extract_info,
                    request.url,
                    real_ydl_opts,
                    5,
                    _YOUTUBE_DL_PROXY_LIST)
            except Exception as e:
                log.exception(e)
                await context.abort(
                    grpc.StatusCode.INTERNAL,
                    "unknown error occurred while extracting info",
                )

            json_info = json.dumps(info)

            if self.cache is not None:
                await self.cache.set(
                    await gen_key(request.url, json.dumps(ydl_opts)),
                    json_info
                )

            # FIXME: Yes, this is a hack.
            # youtube_dl sometimes returns tuples for some repeated fields
            # and ParseDict doesn't like that.
            info = json.loads(json_info)

        try:
            entries = info['entries'][0]['entries']
            for entry in entries:
                yield ParseDict(
                    entry,
                    ExtractInfoResponse(),
                    ignore_unknown_fields=True)
        except KeyError:
            try:
                entries = info['entries']
                for entry in entries:
                    yield ParseDict(
                        entry,
                        ExtractInfoResponse(),
                        ignore_unknown_fields=True)
            except KeyError:
                yield ParseDict(
                    info,
                    ExtractInfoResponse(),
                    ignore_unknown_fields=True)
